from pathlib import Path

import pyexiv2
from PIL import Image


def get_root():
    return Path(__file__).parent.absolute()


def parse_resolution(resolution: str) -> tuple[int, int]:
    """Parse resolution string to tuple of integers.

    Args:
        resolution (str): Resolution string in the format 'width x height'.

    Raises:
        ValueError: If the resolution string is invalid.

    Returns:
        tuple[int, int]: Tuple of integers representing the resolution.
    """
    try:
        return tuple(map(int, resolution.split("x")))
    except ValueError:
        raise ValueError("Invalid resolution format. Use 'width x height'.")


def prune_by_resolution(
    input_file: str | Path, resolution: tuple[int, int], verbose: bool = False
) -> bool:
    size = (0, 0)
    with Image.open(input_file) as im:
        size = im.size

    if size < resolution:
        input_file.unlink()
        if verbose:
            print(f"Removed {input_file}, resolution: {size} < {resolution}")
        return True
    return False


def get_appdata_dir(path_under=None):
    if path_under:
        return Path.home().joinpath("AppData", "Local", "pinterest-dl", path_under)
    return Path.home().joinpath("AppData", "Local", "pinterest-dl")


def write_img_comment(image_path, comment):
    with pyexiv2.Image(str(image_path)) as img:
        img.modify_exif({"Exif.Image.XPComment": comment})

def write_img_subject(image_path, subject):
    with pyexiv2.Image(str(image_path)) as img:
        img.modify_exif({"Exif.Image.XPSubject": subject})



import concurrent.futures
from pathlib import Path
from typing import Literal

import requests
from tqdm import tqdm


def fetch(url: str, response_format: Literal["json", "text"] = "text"):
    if isinstance(url, str):
        req = requests.get(url)
        req.raise_for_status()
        if response_format == "json":
            return req.json()
        elif response_format == "text":
            return req.text
    else:
        print("URL must be a string.")


def download(url: str, output_dir, chunk_size=2048):
    if isinstance(url, str):
        req = requests.get(url)
        req.raise_for_status()

        filename = Path(url).name
        outfile = Path.joinpath(Path(output_dir), filename)
        # create directory if not exist
        outfile.parent.mkdir(parents=True, exist_ok=True)
        with open(outfile, "wb") as payload:
            for chunk in req.iter_content(chunk_size):
                payload.write(chunk)
        return outfile
    else:
        print("URL must be a string.")


def download_with_fallback(url: str, output_dir, fallback_url, chunk_size=2048):
    try:
        return download(url, output_dir, chunk_size)
    except requests.exceptions.HTTPError:
        return download(fallback_url, output_dir, chunk_size)


def download_concurrent(urls: list, output_dir, chunk_size=2048, verbose=False):
    results = [None] * len(urls)  # Initialize a list to hold the results in order
    with concurrent.futures.ThreadPoolExecutor() as executor, tqdm(
        total=len(urls), desc="Downloading"
    ) as pbar:
        futures = {
            executor.submit(download, url, output_dir, chunk_size): idx
            for idx, url in enumerate(urls)
        }
        for future in concurrent.futures.as_completed(futures):
            result_index = futures[future]  # Get the original index for the result
            outfile = future.result()
            results[result_index] = outfile  # Place the result in the corresponding position
            pbar.update(1)
    return results


def download_concurrent_with_fallback(
    urls: list, output_dir, fallback_urls, chunk_size=2048, verbose=False
):
    results = [None] * len(urls)  # Initialize a list to hold the results in order
    with concurrent.futures.ThreadPoolExecutor() as executor, tqdm(
        total=len(urls), desc="Downloading"
    ) as pbar:
        futures = {
            executor.submit(download_with_fallback, url, output_dir, fallback_url, chunk_size): idx
            for idx, (url, fallback_url) in enumerate(zip(urls, fallback_urls))
        }
        for future in concurrent.futures.as_completed(futures):
            result_index = futures[future]  # Get the original index for the result
            outfile = future.result()
            results[result_index] = outfile  # Place the result in the corresponding position
            pbar.update(1)
    return results



import argparse

from pinterest_dl import __description__, __version__


# fmt: off
def get_parser():
    parser = argparse.ArgumentParser(description=__description__ + " v" + __version__)
    parser.add_argument("-v", "--version", action="version", version="v"+__version__)

    cmd = parser.add_subparsers(dest="cmd", help="Command to run")

    # scrape command
    scrape_cmd = cmd.add_parser("scrape", help="Scrape images from Pinterest")
    scrape_cmd.add_argument("url", help="URL to scrape images from")
    scrape_cmd.add_argument("output", help="Output directory")
    scrape_cmd.add_argument("-l", "--limit", type=int, default=100, help="Max number of image to scrape (default: 100)")
    scrape_cmd.add_argument("-r", "--resolution", type=str, help="Minimum resolution to keep (e.g. 512x512).")
    scrape_cmd.add_argument("--timeout", type=int, default=3, help="Timeout in seconds for requests (default: 3)")
    scrape_cmd.add_argument("--incognito", action="store_true", help="Incognito mode")
    scrape_cmd.add_argument("--json", action="store_true", help="Write urls to json file")
    scrape_cmd.add_argument("--dry-run", action="store_true", help="Run without download")
    scrape_cmd.add_argument("--firefox", action="store_true", help="Use Firefox browser")
    scrape_cmd.add_argument("--headful", action="store_true", help="Run in headful mode with browser window")
    scrape_cmd.add_argument("--verbose", action="store_true", help="Print verbose output")

    # download command
    download_cmd = cmd.add_parser("download", help="Download images")
    download_cmd.add_argument("input", help="Input json file containing image urls")
    download_cmd.add_argument("-o", "--output", help="Output directory (default: ./<json_filename>)")
    download_cmd.add_argument("-r", "--resolution", type=str, help="minimum resolution to keep (e.g. 512x512).")
    download_cmd.add_argument("--verbose", action="store_true", help="Print verbose output")

    return parser
# fmt: on



from pathlib import Path

from pinterest_dl import api, cli_parser, io, utils


def main():
    parser = cli_parser.get_parser()
    args = parser.parse_args()

    if args.cmd == "scrape":
        api.run_scrape(
            args.url,
            args.limit,
            args.output,
            timeout=args.timeout,
            json=args.json,
            firefox=args.firefox,
            incognito=args.incognito,
            headful=args.headful,
            dry_run=args.dry_run,
            verbose=args.verbose,
            min_resolution=utils.parse_resolution(args.resolution) if args.resolution else None,
        )
        print("\nDone.")
    elif args.cmd == "download":
        # prepare image url data
        img_datas = io.read_json(args.input)
        srcs, alts, fallbacks, origins = [], [], [], []
        for img_data in img_datas:
            srcs.append(img_data["src"])
            alts.append(img_data["alt"])
            fallbacks.append(img_data["fallback"])
            origins.append(img_data["origin"])

        output_dir = args.output or str(Path(args.input).stem)

        # download images
        downloaded_files = api.run_download(srcs, fallbacks, output_dir, args.verbose)

        # post process
        pruned_idx = api.run_prune(downloaded_files, args.resolution)
        api.run_caption(downloaded_files, alts, origins, pruned_idx, verbose=args.verbose)
        print("\nDone.")
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



import json
import os
import zipfile
from typing import List


def append_json(data, filename, indent=None):
    with open(filename, "r+") as f:
        file_data = json.load(f)
        file_data.update(data)
        f.seek(0)
        json.dump(file_data, f, indent=indent)


def write_json(data, filename, indent=None):
    with open(filename, "w") as f:
        json.dump(data, f, indent=indent)


def read_json(filename):
    with open(filename, "r") as f:
        return json.load(f)


def write_text(data: str | List[str], filename: str):
    if isinstance(data, list):
        data = "\n".join(data)
    with open(filename, "w") as f:
        f.write(data)


def unzip(zip_path, extract_to, target_file=None, verbose=False):
    """
    Extract a specific file from a zip file if target_file is specified.
    Extracts everything otherwise.

    Parameters:
    zip_path (str): Path to the zip file.
    extract_to (str): Directory where the files should be extracted to.
    target_file (str, optional): Specific file to extract. If None, everything is extracted.
    """
    if not zip_path or not str(zip_path).endswith(".zip"):
        raise ValueError(f"Invalid zip file. [{zip_path}]")
    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        if target_file:
            # Target file specified, extract only this file
            for file in zip_ref.namelist():
                if file.endswith(target_file):
                    zip_ref.extract(file, extract_to)
                    # Move the file if it's within a directory
                    extracted_path = os.path.join(extract_to, file)
                    final_path = os.path.join(extract_to, os.path.basename(target_file))
                    if os.path.exists(final_path):
                        os.remove(final_path)
                    os.rename(extracted_path, final_path)
                    # Attempt to remove the directory structure if any
                    dir_path = os.path.dirname(extracted_path)
                    if dir_path != extract_to:  # Check to avoid deleting the extract_to dir
                        os.removedirs(dir_path)
                    if verbose:
                        print(f"{target_file} has been extracted to {final_path}")
                    break
            else:
                print(f"{target_file} was not found in the zip file.")
        else:
            # No specific file to extract, extract everything
            zip_ref.extractall(extract_to)
            print(f"All files have been extracted to {extract_to}")



import os
import platform
import subprocess
import sys
from pathlib import Path
from typing import Literal

from pinterest_dl import downloader, io


class ChromeDriverInstaller:
    def __init__(self, install_dir: Path | str):
        self.install_dir = Path(install_dir)
        self.chrome_version = self.get_chrome_version()
        self.platform = self.get_platform()

    def get_chrome_version(self) -> str:
        try:
            # Determine the operating system
            os_type = platform.system()

            if os_type == "Windows":
                # Command for Windows to read from registry
                output = subprocess.check_output(
                    r'reg query "HKEY_CURRENT_USER\Software\Google\Chrome\BLBeacon" /v version',
                    shell=True,
                )
                version = output.decode().split()[-1]

            elif os_type == "Darwin":  # macOS
                # Command for macOS to get version from Chrome app
                output = subprocess.check_output(
                    ["/Applications/Google Chrome.app/Contents/MacOS/Google Chrome", "--version"]
                )
                version = output.decode("utf-8").strip().split()[-1]

            elif os_type == "Linux":
                # Command for Linux to get version from installed Chrome
                output = subprocess.check_output(["google-chrome", "--version"])
                version = output.decode("utf-8").strip().split()[-1]

            else:
                return "Unsupported operating system."

            return version

        except Exception as e:
            return f"Error retrieving Chrome version: {e}"

    def get_platform(self) -> str:
        os_name = platform.system()
        arch = platform.machine()

        if os_name == "Windows":
            if sys.maxsize > 2**32:
                return "win64"  # 64-bit Windows
            else:
                return "win32"  # 32-bit Windows
        elif os_name == "Darwin":
            if arch == "x86_64":
                return "mac-x64"  # Intel Mac
            elif arch == "arm64":
                return "mac-arm64"  # Apple Silicon Mac
        elif os_name == "Linux":
            if arch in ("x86_64", "amd64"):
                return "linux64"  # 64-bit Linux

        return None

    def install(
        self,
        version: str = "latest",
        platform: Literal["win32", "win64", "mac-x64", "mac-arm64", "linux64", "auto"] = "auto",
        verbose: bool = False,
    ):
        """Install Chrome driver to the specified directory.

        Args:
            install_dir (Path | str): Directory to install Chrome driver.
            version (str): Version of Chrome driver to install. Defaults to 'latest'.
            platform (Literal[&quot;win64&quot;, &quot;win32&quot;], optional): Platform of Chrome driver to install. Defaults to &quot;win64&quot;.
            verbose (bool, optional): Print debug logs. Defaults to False.

        Raises:
            ValueError: _description_
        """
        if version == "latest":
            response = downloader.fetch(
                "https://googlechromelabs.github.io/chrome-for-testing/last-known-good-versions.json",
                response_format="json",
            )
            version = response["channels"]["Stable"]["version"]
        version.strip()

        if platform == "auto":
            platform = self.platform
            if verbose:
                print(f"Auto detected platform: {platform}")

        platform.strip().lower()
        if platform not in ["win32", "win64", "mac-x64", "mac-arm64", "linux64"]:
            raise ValueError(
                "Platform must be one of 'win32', 'win64', 'mac-x64', 'mac-arm64', 'linux64'."
            )
        # create directory if not exist
        Path(self.install_dir).mkdir(parents=True, exist_ok=True)

        url = f"https://storage.googleapis.com/chrome-for-testing-public/{version}/{platform}/chromedriver-{platform}.zip"
        if verbose:
            print(f"Downloading Chrome driver from {url}")
        zip_file = downloader.download(url, self.install_dir)
        io.unzip(zip_file, self.install_dir, "chromedriver.exe", verbose=verbose)
        io.write_text(version, self.install_dir / "CHROMEDRIVER_VERSION")
        print("Chrome driver installed.")

        os.unlink(zip_file)
        if verbose:
            print("Clean up zip file.")



from pathlib import Path
from typing import List, Tuple

from tqdm import tqdm

from pinterest_dl import downloader, io, scraper, utils


def run_download(
    urls: List[str], fallbacks: List[str], output: str | Path, verbose: bool = False
) -> List[str | Path]:
    """download images concurrently with fallback urls

    Args:
        img_urls (List[str]): list of image urls
        fallbacks (List[str]): list of fallback urls
        output (str | Path): output directory to save images
        verbose (bool, optional): print debug logs. Defaults to False.

    Returns:
        List[str  |  Path]: list of downloaded image paths
    """
    return downloader.download_concurrent_with_fallback(urls, output, fallbacks, verbose=verbose)


def run_caption(
    files: List[str | Path],
    captions: List[str],
    origins: List[str],
    indexs: List[int],
    verbose: bool = False,
):
    """write captions to image files

    Args:
        files (List[str  |  Path]): list of image paths
        captions (List[str]): list of captions
        origins (List[str]): list of image origin urls
        indexs (List[int]): list of indexs to write caption
        verbose (bool, optional): print debug logs. Defaults to False.
    """
    for index in indexs:
        try:
            file = files[index]
            caption = captions[index]
            origin = origins[index]
            utils.write_img_comment(file, origin)
            utils.write_img_subject(file, caption)

            if verbose:
                print(f"{file} -> {caption}")
        except Exception as e:
            print(f"Error captioning {file}: {e}")


def run_prune(
    local_images: List[str | Path], min_resolution: Tuple[int, int], verbose: bool = False
):
    """prune images by resolution

    Args:
        local_images (List[str  |  Path]): list of image paths
        min_resolution (Tuple[int, int]): minimum resolution to keep
    """
    new_index = []
    if min_resolution:
        for index, img in tqdm(enumerate(local_images), desc="Pruning"):
            if not utils.prune_by_resolution(img, min_resolution, verbose=verbose):
                new_index.append(index)

    else:
        new_index = range(len(local_images))
    print(f"Pruned {len(local_images) - len(new_index)} images.")
    return new_index


def run_scrape(
    url: str,
    limit: int,
    output: str | Path,
    timeout: int = 3,
    json: bool = False,
    json_path: str | Path = None,
    firefox: bool = False,
    incognito: bool = False,
    headful: bool = True,
    dry_run: bool = False,
    verbose: bool = False,
    min_resolution: Tuple[int, int] = None,
    caption: bool = True,
):
    """extracts image urls from pinterest board and downloads them

    Args:
        url (str): input pinterest board url
        limit (int): max number of image to scrape
        output (str | Path): output directory to save images
        persistence (int, optional): time to wait for page load. Defaults to 120.
        json (bool): write image urls to json file. Defaults to False.
        json_path (str | Path): path to write json file. Defaults to None.
        firefox (bool, optional): use firefox browser. Defaults to False.
        incognito (bool, optional): use incognito mode. Defaults to False.
        headful (bool, optional): run in headful mode with browser window. Defaults to True.
        dry_run (bool, optional): only print image urls. Defaults to False.
        verbose (bool, optional): print debug logs. Defaults to False.
        min_resolution (Tuple[int, int], optional): minimum resolution to keep. Defaults to None.
        caption (bool, optional): write image caption as metadata. Defaults to True.
    """
    if firefox:
        browser = scraper.Browser().Firefox(incognito=incognito, headful=headful)
    else:
        browser = scraper.Browser().Chrome(
            exe_path=utils.get_appdata_dir("chromedriver.exe"),
            incognito=incognito,
            headful=headful,
        )

    try:
        pin_scraper = scraper.Pinterest(browser)
        imgs = pin_scraper.scrape(
            url,
            limit=limit,
            timeout=timeout,
            verbose=verbose,
        )
        srcs, alts, fallbacks, origins = [], [], [], []
        for i in imgs:
            srcs.append(i["src"])
            alts.append(i["alt"])
            fallbacks.append(i["fallback"])
            origins.append(i["origin"])
    finally:
        browser.close()

    if json:
        output_path = json_path or str(Path(output).absolute().name) + ".json"
        if json_path:
            output_path = json_path
        io.write_json(imgs, output_path, indent=4)
    if not dry_run:
        downloaded_files = run_download(srcs, fallbacks, output, verbose)
        # post download
        pruned_idx = run_prune(downloaded_files, min_resolution)
        if caption:
            run_caption(downloaded_files, alts, origins, pruned_idx, verbose=verbose)
    else:
        for i in imgs:
            print(i)



import copy
import os
import random
import socket
import time
from pathlib import Path

from selenium import webdriver
from selenium.common.exceptions import StaleElementReferenceException
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement
from tqdm import tqdm

from pinterest_dl import utils
from pinterest_dl.driver_installer import ChromeDriverInstaller


def randdelay(a, b):
    time.sleep(random.uniform(a, b))


class BrowserVersion:
    def __init__(self, major=0, minor=0, build=0, patch=0):
        self.Major: int = major
        self.Minor: int = minor
        self.Build: int = build
        self.Patch: int = patch

    @staticmethod
    def from_str(version: str) -> "BrowserVersion":
        segs = version.split(".")
        if len(segs) != 4:
            raise ValueError(
                "Invalid version string. Must be in the format 'major.minor.build.patch'"
            )
        return BrowserVersion(int(segs[0]), int(segs[1]), int(segs[2]), int(segs[3]))

    def __str__(self):
        return f"{self.Major}.{self.Minor}.{self.Build}.{self.Patch}"


class Browser:
    def __init__(self):
        self.app_root = utils.get_appdata_dir()
        self.version: BrowserVersion = None

    def _validate_chrome_driver_version(self):
        version_file = Path(self.app_root, "CHROMEDRIVER_VERSION")
        if not version_file.exists():
            return False

        with open(version_file, "r") as f:
            version_str = f.read().strip()
            current_version = BrowserVersion.from_str(version_str)
        print(f"Current Chrome driver version: {current_version}")
        if self.version.Major != current_version.Major:
            return False
        if self.version.Minor != current_version.Minor:
            return False
        if self.version.Build != current_version.Build:
            return False
        # patch version can be different

        return True

    def Chrome(
        self, image_enable=False, incognito=False, exe_path="chromedriver.exe", headful=False
    ):
        driver_installer = ChromeDriverInstaller(self.app_root)
        self.version = BrowserVersion.from_str(driver_installer.chrome_version)

        if not os.path.exists(exe_path) or not self._validate_chrome_driver_version():
            print(f"Installing latest Chrome driver for version {self.version}")
            driver_installer.install(self.app_root, version="latest", platform="auto")

        service = webdriver.chrome.service.Service(exe_path)
        chrome_options = webdriver.ChromeOptions()

        # Disable images
        # https://scrapeops.io/selenium-web-scraping-playbook/python-selenium-disable-image-loading/
        chrome_options.add_argument(
            "--blink-settings=imagesEnabled=true"
            if image_enable
            else "--blink-settings=imagesEnabled=false"
        )
        chrome_options.add_argument("--log-level=3")  # Suppress most logs
        if incognito:
            print("Running in incognito mode")
            chrome_options.add_argument("--incognito")
        if not headful:
            print("Running in headless mode")
            chrome_options.add_argument("--headless=new")
        browser = webdriver.Chrome(options=chrome_options, service=service)
        return browser

    def Firefox(self, image_enable=False, incognito=False, headful=False):
        firefox_options = webdriver.FirefoxOptions()
        # Disable images
        if image_enable:
            firefox_options.set_preference("permissions.default.image", 1)
        else:
            firefox_options.set_preference("permissions.default.image", 2)
        if incognito:
            firefox_options.set_preference("browser.privatebrowsing.autostart", True)
        if not headful:
            print("Running in headless mode")
            firefox_options.add_argument("--headless")
        browser = webdriver.Firefox(options=firefox_options)
        return browser


class Pinterest:
    def __init__(self, browser=None):
        self.browser: WebDriver = browser

    # currently not used
    def login(self, email, password):
        self.browser.get("https://www.pinterest.com.au/login/")
        email_field = self.browser.find_element(By.ID, "email")
        email_field.send_keys(email)
        password_field = self.browser.find_element(By.ID, "password")
        password_field.send_keys(password)
        randdelay(1, 2)  # delay between 1 and 2 seconds
        password_field.send_keys(Keys.RETURN)

    def scrape(
        self,
        url,
        limit=20,
        timeout=3,
        verbose=False,
    ):
        unique_results = set()  # Use a set to store unique results
        imgs_data = []  # Store image data
        previous_divs = []
        tries = 0
        pbar = tqdm(total=limit, desc="Scraping")
        try:
            self.browser.get(url)
            while len(unique_results) < limit:
                try:
                    divs = self.browser.find_elements(By.CSS_SELECTOR, "div[data-test-id='pin']")
                    if divs == previous_divs:
                        tries += 1
                        time.sleep(1)  # delay 1 second
                    else:
                        tries = 0
                    if tries > timeout:
                        print(f"\nTimeout: no new images in ({timeout}) seconds.")
                        break

                    for div in divs:
                        if self._is_div_ad(div) or len(unique_results) >= limit:
                            continue
                        images = div.find_elements(By.TAG_NAME, "img")
                        href = div.find_element(By.TAG_NAME, "a").get_attribute("href")
                        for image in images:
                            alt = image.get_attribute("alt")
                            src = image.get_attribute("src")
                            if src and "/236x/" in src:
                                src = src.replace("/236x/", "/originals/")
                                src_736 = src.replace("/originals/", "/736x/")
                                if src not in unique_results:
                                    unique_results.add(src)
                                    img_data = {
                                        "src": src,
                                        "alt": alt,
                                        "fallback": src_736,
                                        "origin": href,
                                    }
                                    imgs_data.append(img_data)
                                    pbar.update(1)
                                    if verbose:
                                        print(src, alt)
                                    if len(unique_results) >= limit:
                                        break

                    previous_divs = copy.copy(divs)  # copy to avoid reference

                    # Scroll down
                    dummy = self.browser.find_element(By.TAG_NAME, "a")
                    dummy.send_keys(Keys.PAGE_DOWN)
                    randdelay(1, 2)  # delay between 1 and 2 seconds

                except StaleElementReferenceException:
                    if verbose:
                        print("\nStaleElementReferenceException")

        except (socket.error, socket.timeout):
            print("Socket Error")
        finally:
            pbar.close()
            if verbose:
                print(f"Scraped {len(imgs_data)} images")
            return imgs_data

    def _is_div_ad(self, div: WebElement):
        """Check if div is an ad.

        Args:
            div (webElement): div element.
        """
        ads_svg_path = "M12 9a3 3 0 1 0 0 6 3 3 0 0 0 0-6M3 9a3 3 0 1 0 0 6 3 3 0 0 0 0-6m18 0a3 3 0 1 0 0 6 3 3 0 0 0 0-6"
        svg_elements = div.find_elements(By.TAG_NAME, "svg")
        for svg in svg_elements:
            if ads_svg_path in svg.get_attribute("innerHTML"):
                return True
